/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.common;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;

import org.objectweb.petals.component.common.bc.ExternalListenerManager;
import org.objectweb.petals.component.common.listener.ComponentConfiguration;
import org.objectweb.petals.component.common.listener.JBIListener;
import org.objectweb.petals.component.common.listener.JBIListenerManager;
import org.objectweb.petals.component.common.sender.MessageExchangeSender;
import org.objectweb.petals.component.common.su.ServiceUnitExtensionsValidator;
import org.objectweb.petals.component.common.su.ServiceUnitListener;
import org.objectweb.petals.component.common.su.SimpleServiceUnitManager;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

/**
 * An abstract class that can be used to easily create a new JBI Component. It
 * implements Component, ComponentLifeCycle and MessageExchangeProcessor JBI
 * interfaces.
 *
 * @author ofabre
 *
 */
public abstract class AbstractComponent implements Component,
		ComponentLifeCycle {

	/**
	 * An initialization class that handles the JBIListener and the
	 * ExternalListenerManager
	 */
	private ComponentInitializer initializer;

	/**
	 * The component's delivery channel
	 */
	private DeliveryChannel channel;

	/**
	 * The Component's Context
	 */
	private ComponentContext context;

	/**
	 * The extension properties. The property file has been created during the
	 * bootstrap phase
	 */
	protected ComponentPropertiesManager extensionProperties;

	/**
	 * The message exchange listener
	 */
	private JBIListenerManager listenerManager;

	/**
	 * The component's logger
	 */
	private Logger log; // NOPMD by ofabre

	/**
	 * The ServiceUnit lifecycle listener
	 */
	private ServiceUnitListener serviceUnitListener;

	/**
	 * The service unit extensions validator
	 */
	private ServiceUnitExtensionsValidator serviceUnitExtensionsValidator;

	/**
	 * The component's service unit manager
	 */
	private SimpleServiceUnitManager suManager;

	/**
	 * A useful class wrapping message exchange sending
	 */
	private MessageExchangeSender sender;

	/**
	 * Creates a new {@link JBIListenerManager}
	 *
	 * @param configuration
	 *            Configuration of the component
	 * @return the created {@link JBIListenerManager}
	 */
	private JBIListenerManager createJBIListenerManager(
			final ComponentConfiguration configuration, JBIListener listener) {
		return new JBIListenerManager(channel, configuration, context
				.getComponentName(), log, suManager, listener);
	}

	/**
	 * Can be overidden to provide a specific service unit manager
	 *
	 * @return the service unit manager used by this component.
	 */
	protected SimpleServiceUnitManager createServiceUnitManager() {
		return new SimpleServiceUnitManager(this, context, log);
	}

	/**
	 * Can be overidden to provide a specific way to configure the thread pool.
	 *
	 * @return the thread pool configuration.
	 */
	protected ComponentConfiguration createComponentConfiguration() {
		ComponentConfiguration conf = new ComponentConfiguration();
		conf.setPoolSize(Integer.parseInt(extensionProperties.getProperty(
				ComponentPropertiesManager.POOL_SIZE, "10")));

		conf.setIgnoredStatus(IgnoredStatusConstants
				.valueOf(extensionProperties.getProperty(
						ComponentPropertiesManager.IGNORED_STATUS,
						ComponentPropertiesManager.DEFAULT_IGNORED_STATUS)));

		this.log.info("Component configuration -> Thread pool size : "
				+ conf.getPoolSize() + ", Ignored message exchange status : "
				+ conf.getIgnoredStatus().value());

		return conf;
	}

	/**
	 * Returns the component delivery channel, used to send messages to petals
	 * container
	 *
	 * @return the delivery channel. Not null.
	 */
	public DeliveryChannel getChannel() {
		return channel;
	}

	/**
	 * The component context
	 *
	 * @return the component context. Not null.
	 */
	public ComponentContext getContext() {
		return context;
	}

	/**
	 * Returns null by default as this class does not provides JMX extension
	 * MBean
	 *
	 * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
	 */
	public ObjectName getExtensionMBeanName() {
		return null;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.Component#getLifeCycle()
	 */
	public ComponentLifeCycle getLifeCycle() {
		return this;
	}

	/**
	 * Returns the component logger
	 *
	 * @return component logger. Not null.
	 */
	public Logger getLogger() {
		return log;
	}

	/**
	 * Can be overidden to return a custom resource bundle name
	 *
	 * @return the resource bundle name or null if no resource bundle name has
	 *         to be used (default value)
	 */
	protected String getResourceBundleName() {
		return null;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.Component#getServiceDescription(javax.jbi.servicedesc.ServiceEndpoint)
	 */
	public Document getServiceDescription(ServiceEndpoint endpoint) {
		return suManager.getServiceDescription(endpoint);
	}

	public ServiceUnitListener getServiceUnitListener() {
		return serviceUnitListener;
	}

	public ServiceUnitExtensionsValidator getServiceUnitExtensionsValidator() {
		return serviceUnitExtensionsValidator;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.Component#getServiceUnitManager()
	 */
	public ServiceUnitManager getServiceUnitManager() {
		return suManager;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.ComponentLifeCycle#init(javax.jbi.component.ComponentContext)
	 */
	public void init(ComponentContext compContext) throws JBIException {
		this.context = compContext;
		this.log = context.getLogger("", getResourceBundleName());
		log.fine("Component initialization in progress...");
		channel = context.getDeliveryChannel();
		suManager = createServiceUnitManager();
		extensionProperties = loadProperties();
		initializer = new ComponentInitializer();
		log.fine("Calling specific initialization...");
		init(initializer);
		log.fine("Specific initialization done.");
		sender = createMessageExchangeSender();
		log.log(Level.INFO, "Component initialized");
	}

	private MessageExchangeSender createMessageExchangeSender() {
		return new MessageExchangeSender(getChannel(), getContext(), suManager);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.Component#isExchangeWithConsumerOkay(javax.jbi.servicedesc.ServiceEndpoint,
	 *      javax.jbi.messaging.MessageExchange)
	 */
	public boolean isExchangeWithConsumerOkay(ServiceEndpoint endpoint,
			MessageExchange exchange) {
		return true;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.Component#isExchangeWithProviderOkay(javax.jbi.servicedesc.ServiceEndpoint,
	 *      javax.jbi.messaging.MessageExchange)
	 */
	public boolean isExchangeWithProviderOkay(ServiceEndpoint endpoint,
			MessageExchange exchange) {
		return true;
	}

	/**
	 * Load the extensions properties
	 */
	protected ComponentPropertiesManager loadProperties() {
		ComponentPropertiesManager extensionProperties = new ComponentPropertiesManager();
		extensionProperties.setRootPath(getContext().getInstallRoot());
		extensionProperties.load();
		return extensionProperties;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.Component#resolveEndpointReference(org.w3c.dom.DocumentFragment)
	 */
	public ServiceEndpoint resolveEndpointReference(DocumentFragment epr) {
		return null;
	}

	public void setServiceUnitListener(ServiceUnitListener serviceUnitListener) {
		this.serviceUnitListener = serviceUnitListener;
	}

	public void setServiceUnitExtensionsValidator(
			ServiceUnitExtensionsValidator serviceUnitExtensionsValidator) {
		this.serviceUnitExtensionsValidator = serviceUnitExtensionsValidator;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.ComponentLifeCycle#shutDown()
	 */
	public void shutDown() throws JBIException {
		log.log(Level.INFO, "shutdown");
		listenerManager = null; // NOPMD frees the memory while component is
		// shutdown
		context = null; // NOPMD idem
		log = null; // NOPMD idem
		channel = null; // NOPMD idem
		suManager = null; // NOPMD idem
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.ComponentLifeCycle#start()
	 */
	public void start() throws JBIException {
		log.log(Level.INFO, "start");
		ComponentConfiguration configuration = createComponentConfiguration();
		listenerManager = createJBIListenerManager(configuration, initializer
				.getJbiListener());
		listenerManager.startListening();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see javax.jbi.component.ComponentLifeCycle#stop()
	 */
	public void stop() throws JBIException {
		log.log(Level.INFO, "stop");
		// FIXME : call terminate or close?
		listenerManager.stopListening();
	}

	/**
	 * This method must be overidden to initialize JBIListener and
	 * ExternalListenerManager handled by the {@link ComponentInitializer}.
	 * This method is called at the end of the component classical
	 * initialization. When this method is called, the component's delivery
	 * channel, context, logger and service unit manager are available.<br/>
	 *
	 * Service Engine case : <br/> You must provide your custom
	 * {@link JBIListener} by calling "setJbiListener" on the
	 * {@link ComponentInitializer}. {@link ExternalListenerManager} isn't
	 * required in this case
	 *
	 * Binding Component case : <br/> If this binding component must handle jbi
	 * messages, you must provide your custom {@link JBIListener} by calling
	 * "setJbiListener" on the BindingComponentInitializer.<br/> If this
	 * binding component must handle messages from external service to internal
	 * jbi components, you must provide your custom
	 * {@link ExternalListenerManager} by calling "setExternalListenerManager"
	 * on the {@link ComponentInitializer}.
	 *
	 * @param initializer
	 *            the {@link ComponentInitializer} to fill
	 */
	protected abstract void init(ComponentInitializer initializer)
			throws JBIException;

	/**
	 * An initialization class that handles the JBIListener and/or the
	 * ExternalListenerManager
	 *
	 * @author ofabre
	 *
	 */
	public class ComponentInitializer {

		private JBIListener jbiListener;

		private ExternalListenerManager externalListenerManager;

		public ExternalListenerManager getExternalListenerManager() {
			return externalListenerManager;
		}

		public void setExternalListenerManager(
				ExternalListenerManager externalListenerManager) {
			this.externalListenerManager = externalListenerManager;
		}

		public JBIListener getJbiListener() {
			return jbiListener;
		}

		public void setJbiListener(JBIListener jbiListener) {
			this.jbiListener = jbiListener;
		}

	}

	public ComponentInitializer getInitializer() {
		return initializer;
	}

	public MessageExchangeSender getSender() {
		return sender;
	}
}
