/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.common;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.ReflectionException;
import javax.management.RuntimeOperationsException;

/**
 * The MBean used to expose component attributes and operations via JMX.
 *
 * @author Christophe HAMERLING - eBMWebSourcing
 *
 */
public class ExtensionMBean implements DynamicMBean {

    private final static Hashtable<String, Class<?>> primitiveClasses = new Hashtable<String, Class<?>>(
            8);

    private ManagedBootstrap managedBootstrap;
    {
        primitiveClasses.put(Boolean.TYPE.toString(), Boolean.TYPE);
        primitiveClasses.put(Character.TYPE.toString(), Character.TYPE);
        primitiveClasses.put(Byte.TYPE.toString(), Byte.TYPE);
        primitiveClasses.put(Short.TYPE.toString(), Short.TYPE);
        primitiveClasses.put(Integer.TYPE.toString(), Integer.TYPE);
        primitiveClasses.put(Long.TYPE.toString(), Long.TYPE);
        primitiveClasses.put(Float.TYPE.toString(), Float.TYPE);
        primitiveClasses.put(Double.TYPE.toString(), Double.TYPE);
    }

    /**
     * Creates a new instance of {@link ExtensionMBean}
     *
     * @param bootstrap
     */
    public ExtensionMBean(ManagedBootstrap bootstrap) {
        this.managedBootstrap = bootstrap;
    }

    /**
     * Find the classes of the signature.
     *
     * @param signature
     * @param loader
     * @return
     * @throws ReflectionException
     */
    private Class[] findSignatureClasses(String signature[], ClassLoader loader)
            throws ReflectionException {
        if (signature == null) {
            return null;
        }

        final int length = signature.length;
        final Class tab[] = new Class[length];

        if (length == 0) {
            return tab;
        }

        try {
            for (int i = 0; i < length; i++) {

                // try to get the primitive class
                final Class primCla = (Class) ExtensionMBean.primitiveClasses
                        .get(signature[i]);
                if (primCla != null) {
                    tab[i] = primCla;
                } else {
                    // try to get the class if not primitive one
                    tab[i] = Class.forName(signature[i], false, loader);
                }
            }
        } catch (ClassNotFoundException e) {
            throw new ReflectionException(e,
                    "The parameter class could not be found");
        } catch (RuntimeException e) {
            throw e;
        }
        return tab;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.management.DynamicMBean#getAttribute(java.lang.String)
     */
    public synchronized Object getAttribute(String attribute)
            throws AttributeNotFoundException, MBeanException,
            ReflectionException {

        String value = managedBootstrap.getComponentPropertiesManager()
                .getProperties().getProperty(attribute);
        if (value != null
                && !value
                        .startsWith(ComponentPropertiesManager.MANAGED_METHOD_PREFIX)) {
            return value;
        } else {
            throw new AttributeNotFoundException("Attribute '" + attribute
                    + "' does not exists");
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.management.DynamicMBean#getAttributes(java.lang.String[])
     */
    public synchronized AttributeList getAttributes(String[] attributes) {
        AttributeList list = new AttributeList();
        for (String name : attributes) {
            String value = managedBootstrap.getComponentPropertiesManager()
                    .getProperties().getProperty(name);
            if (value != null
                    && !value
                            .startsWith(ComponentPropertiesManager.MANAGED_METHOD_PREFIX)) {
                list.add(new Attribute(name, value));
            }
        }
        return list;
    }

    /**
     * Dynamically get attributes for the MBean. The attributes are the ones
     * from the properties file.
     *
     * @return
     */
    protected MBeanAttributeInfo[] getAttributesInfos() {
        // get the attributes names as sorted strings
        Set<String> attributeKeys = managedBootstrap
                .getComponentPropertiesManager().getExposedAttributeKeys();

        // build the MBeanAttributeInfo
        MBeanAttributeInfo[] attrs = new MBeanAttributeInfo[attributeKeys
                .size()];
        Iterator<String> it = attributeKeys.iterator();
        for (int i = 0; i < attrs.length; i++) {
            String name = it.next();
            attrs[i] = new MBeanAttributeInfo(name, "java.lang.String",
                    "Property " + name, true, true, false);
        }
        return attrs;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.management.DynamicMBean#getMBeanInfo()
     */
    public synchronized MBeanInfo getMBeanInfo() {
        // get the attributes
        MBeanAttributeInfo[] attrs = getAttributesInfos();
        // get the operations
        MBeanOperationInfo[] opps = getOperationsInfos();

        // create MBeanInfo with no constructors and no notifications
        return new MBeanInfo(this.getClass().getName(),
                "Property Manager MBean", attrs, null, opps, null);
    }

    /**
     * Dynamically get the operations for the MBean. The operations are defined
     * in the properties manager with a special prefix.
     *
     * @return
     */
    protected MBeanOperationInfo[] getOperationsInfos() {
        List<String> methods = managedBootstrap.getComponentPropertiesManager()
                .getExposedMethodNames();
        MBeanOperationInfo[] result = null;

        List<MBeanOperationInfo> operations = new ArrayList<MBeanOperationInfo>();
        Method[] methodList = this.managedBootstrap.getClass().getMethods();

        for (int i = 0; i < methodList.length; i++) {
            Method method = methodList[i];
            String name = method.getName();

            // add method if it has been defined by the user
            if (methods.contains(name)) {
                MBeanOperationInfo oper = new MBeanOperationInfo(
                        "Operation description", method);
                operations.add(oper);
            }
        }

        result = new MBeanOperationInfo[operations.size()];
        operations.toArray(result);

        return result;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.management.DynamicMBean#invoke(java.lang.String,
     *      java.lang.Object[], java.lang.String[])
     */
    public Object invoke(String actionName, Object[] params, String[] signature)
            throws MBeanException, ReflectionException {
        Object result = null;

        final Class objClass = managedBootstrap.getClass();
        final ClassLoader loader = objClass.getClassLoader();
        final Class[] parameters = (signature == null) ? null
                : findSignatureClasses(signature, loader);

        // Query the metadata service to get the right method
        //
        Method method = null;
        try {
            method = objClass.getMethod(actionName, parameters);
        } catch (Exception e) {
            // Nothing to do
        }

        if (method == null) {
            throw new ReflectionException(
                    new NoSuchMethodException(actionName),
                    "The operation with name " + actionName
                            + " could not be found");
        }

        try {
            // invoke method
            result = method.invoke(managedBootstrap, params);
        } catch (IllegalAccessException e) {
            throw new ReflectionException(e, "IllegalAccessException"
                    + " occured trying to invoke operation " + actionName);
        } catch (RuntimeException e) {
            throw new RuntimeOperationsException(e, "RuntimeException"
                    + " occured trying to invoke operation " + actionName);
        } catch (InvocationTargetException e) {
            throw new MBeanException(e, "Exception thrown in operation "
                    + actionName);
        }
        return result;
    }

    /**
     * Save the properties to a file
     *
     * @throws IOException
     */
    private void save() throws IOException {
        managedBootstrap.getComponentPropertiesManager().save();
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.management.DynamicMBean#setAttribute(javax.management.Attribute)
     */
    public synchronized void setAttribute(Attribute attribute)
            throws AttributeNotFoundException, InvalidAttributeValueException,
            MBeanException, ReflectionException {
        String name = attribute.getName();
        if (managedBootstrap.getComponentPropertiesManager().getProperties()
                .getProperty(name) == null) {
            throw new AttributeNotFoundException(name);
        }

        Object value = attribute.getValue();
        if (!(value instanceof String)) {
            throw new InvalidAttributeValueException(
                    "Attribute value not a string: " + value);
        }

        // set the property and save the properties file
        managedBootstrap.getComponentPropertiesManager().getProperties()
                .setProperty(name, (String) value);
        try {
            save();
        } catch (IOException e) {
            throw new MBeanException(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.management.DynamicMBean#setAttributes(javax.management.AttributeList)
     */
    public synchronized AttributeList setAttributes(AttributeList list) {
        AttributeList result = new AttributeList();

        for (Object object : list) {
            Attribute attr = (Attribute) object;
            String name = attr.getName();
            Object value = attr.getValue();
            if (managedBootstrap.getComponentPropertiesManager()
                    .getProperties().getProperty(name) != null
                    && value instanceof String) {
                managedBootstrap.getComponentPropertiesManager()
                        .getProperties().setProperty(name, (String) value);
                result.add(new Attribute(name, value));
            }
        }

        try {
            save();
        } catch (IOException e) {
            return new AttributeList();
        }
        return result;
    }
}
