/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.common.listener;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;

import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.jbi.descriptor.Extensions;

/**
 * This interface is used when a JBI message is received on the Component from
 * the JBI environment. <br>
 * Service Engine case : <br>
 * The goal of this method is to process a service with this JBI message
 * information.<br>
 * For instance, a ServiceUnit has been deployed on the component, defining a
 * <code>provides</code> section.<br>
 * This section defines an Endpoint <em>"myService/jbiMyServiceEndpoint"</em>
 * and some property extensions.<br>
 * A <em>"myService/jbiMyServiceEndpoint"</em> has been activated on the JBI
 * environment.<br>
 * <br>
 * When a JBI consumer sends in the JBI environment a message to the
 * <em>"myService/jbiMyServiceEndpoint"</em> endpoint, this component received
 * the message, and calls the <code>onJBIMessage(exchange, extensions)</code>
 * method.<br>
 * Then, the method has to process the message according to optional property
 * extensions.
 *
 * Binding Component case : <br>
 * The goal of this method is to call an external service, related to this JBI
 * message. The address associated with the requested endpoint is handled by the
 * Extensions object in parameters.<br>
 * For instance, a ServiceUnit has been deployed on the component, defining a
 * <code>provides</code> section.<br>
 * This section defines a relation between the external address
 * <em>"http://www.enterprise.com/webservices/myService"</em> (address
 * attribute of the provides extensions), and an Endpoint
 * <em>"myService/jbiMyServiceEndpoint"</em>.<br>
 * A <em>"myService/jbiMyServiceEndpoint"</em> has been activated on the JBI
 * environment.<br>
 * <br>
 * When a JBI consumer sends in the JBI environment a message to the
 * <em>"myService/jbiMyServiceEndpoint"</em> endpoint, this component received
 * the message, and calls the <code>onJBIMessage(address,endpoint)</code>
 * method.<br>
 * Then, the method has to send the message to
 * <em>"http://www.enterprise.com/webservices/myService"</em>. <br/>
 *
 * @author ofabre - EBM Websourcing
 *
 */
public interface JBIListener {

	/**
	 * <p>
	 * This method is used to process a JBI message received on the Component
	 * from the JBI environment.
	 *
	 * This method must support concurrent invocation for multi-threaded
	 * environments.
	 * </p>
	 *
	 * <p>
	 * This method can't throws Exceptions. You must managed Exception yourself
	 * by using {@link MessageExchangeWrapper} methods :<br/> -
	 * setFault(Exception) : create a {@link Fault} with the given Exception and
	 * set it in the {@link MessageExchange}<br/> - setError(Exception) : set
	 * Error part of the {@link MessageExchange} and set {@link MessageExchange}
	 * status to error. Exceptions handled by {@link MessageExchange}s must be
	 * in the container classpath.
	 * </p>
	 *
	 * <p>
	 * Pattern's use cases :<br/> - InOnly : DONE status is automaticaly set at
	 * the end of the message exchange process.<br/> - RobustInOnly : DONE
	 * status is automaticaly set if no Fault found in exchange at the end of
	 * the message exchange process.<br/> - InOut : DONE status is never set
	 * (forbidden).<br/> - InOptOut : DONE status is automaticaly set if no
	 * Fault or Out message found in exchange at the end of the message exchange
	 * process.
	 *
	 * @param exchange
	 *            the message exchange to process, wrapped to add some util
	 *            methods. Not null
	 * @param extensions
	 *            JBI descriptor extensions. The message exchange targets a
	 *            specific endpoint, this endpoint has been activated during the
	 *            processing of a provides node of a service unit jbi
	 *            descriptor. Theses extensions are the extensions of this
	 *            provides node. Can be null
	 *
	 * @throws RuntimeException
	 *             RuntimeException will be handle by the MessageExchange as
	 *             Error part. So handled {@link RuntimeException}s must be in
	 *             the classpath of your Petals node(s)
	 */
	void onJBIMessage(MessageExchangeWrapper exchange, Extensions extensions);

}
