/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.common.listener;

import java.util.Iterator;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;

import org.objectweb.petals.component.common.su.SimpleServiceUnitManager;

/**
 * This class helps a Component to listen to {@link MessageExchange}s on the
 * DeliveryChannel (accept() method). It acts as a ListenerManager. It starts or
 * stops a pool of thread that listen to new incoming JBI
 * {@link MessageExchange}s<p/>
 * <p>
 * Listener thread pool can be parametrized with few parameters. See
 * constructors.
 * </p>
 *
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis
 */
public class JBIListenerManager {

	private Logger logger; // NOPMD by gblondelle

	private LinkedBlockingQueue<Runnable> workers;

	private ThreadPoolExecutor executor;

	private ComponentConfiguration configuration;

	/**
	 * The component's delivery channel
	 */
	private DeliveryChannel channel;

	/**
	 * The component's service unit manager
	 */
	private SimpleServiceUnitManager suManager;

	private JBIListener listener;

	private String componentName;

	/**
	 *
	 *
	 * @param channel
	 * @param configuration
	 *            The configuration of component
	 * @param componentName
	 *            Name of the component. Used to define the Thread name.
	 * @param log
	 * @param suManager
	 * @param listener
	 */
	public JBIListenerManager(DeliveryChannel channel,
			final ComponentConfiguration configuration, String componentName,
			Logger log, SimpleServiceUnitManager suManager, JBIListener listener) {
		super();
		this.configuration = configuration;
		this.channel = channel;
		this.suManager = suManager;
		this.componentName = componentName;
		this.listener = listener;
		if (log != null) {
			this.logger = log;
		} else {
			this.logger = LogManager.getLogManager().getLogger(componentName);
		}

		this.workers = new LinkedBlockingQueue<Runnable>(configuration
				.getPoolSize());
		this.executor = new ThreadPoolExecutor(configuration.getPoolSize(),
				configuration.getPoolSize(), Long.MAX_VALUE, TimeUnit.SECONDS,
				workers);
	}

	/**
	 * Start threads that listen to {@link MessageExchange}s
	 *
	 */
	public void startListening() {
		for (int i = 0; i < configuration.getPoolSize(); i++) {
			MessageExchangeProcessor processor = new MessageExchangeProcessor(
					channel, logger, suManager, listener, configuration,
					componentName);
			try {
				executor.execute(processor);
			} catch (RejectedExecutionException e) {
				// TODO Treat this exception
				logger.log(Level.SEVERE, e.getMessage(), e);
			}
		}
	}

	/**
	 * Stop threads that listen to {@link MessageExchange}s
	 *
	 */
	public void stopListening() {
		// Stop workers that are working
		for (Iterator iter = executor.getQueue().iterator(); iter.hasNext();) {
			MessageExchangeProcessor element = (MessageExchangeProcessor) iter
					.next();
			element.close();
		}
	}

}
