/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.listener;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.objectweb.petals.component.common.IgnoredStatusConstants;
import org.objectweb.petals.component.common.su.SimpleServiceUnitManager;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.MessageExchangeWrapperImpl;
import org.objectweb.petals.jbi.descriptor.Extensions;
import org.objectweb.petals.jbi.descriptor.Provides;

/**
 * Processor for recevied MessageExchanges.
 *
 * @version $Revision: $ $Date: $
 * @since Petals 1.3
 * @author alouis, ofabre - EBM Websourcing
 */
public class MessageExchangeProcessor extends Thread {

	/**
	 * The component's delivery channel
	 */
	private DeliveryChannel channel;

	/**
	 * The component's logger
	 */
	private Logger log;

	/**
	 * The component's service unit manager
	 */
	private SimpleServiceUnitManager suManager;

	private JBIListener listener;

	private boolean running;

	private ComponentConfiguration configuration;

	/**
	 * Creates a processor that listen to new {@link MessageExchange} on the
	 * specified channel. Exchanges are ignored depending on the specified
	 * exchange-status to ignore. wait for the specified interval between two
	 * 'accept()';
	 *
	 * @param channel
	 * @param log
	 * @param suManager
	 * @param listener
	 * @param configuration
	 *            The configuration of component
	 * @param componentName
	 */
	public MessageExchangeProcessor(DeliveryChannel channel, Logger log,
			SimpleServiceUnitManager suManager, JBIListener listener,
			ComponentConfiguration configuration, String componentName) {
		super(componentName + "-JBI listener thread");
		this.channel = channel;
		this.log = log;
		this.suManager = suManager;
		this.listener = listener;
		this.configuration = configuration;
	}

	/**
	 * Launch the component processing of the given message exchange
	 *
	 * @param exchange
	 *            the message exchange to process. Must be non null.
	 * @throws JBIException
	 *             if an error occured during the message exchange processing
	 */
	public void process(final MessageExchangeWrapper exchange)
			throws MessagingException {
		if (exchange.isActiveStatus()) {
			if (exchange.isProviderRole()) {
				if (exchange.isInOnlyPattern()) {
					processInOnly(exchange);
				} else if (exchange.isRobustInOnlyPattern()) {
					processRobustInOnly(exchange);
				} else if (exchange.isInOutPattern()) {
					processInOut(exchange);
				} else if (exchange.isInOptionalOutPattern()) {
					processInOptionalOut(exchange);
				} else {
					throw new MessagingException(
							"MessageExchangePattern not recognized :"
									+ exchange.getPattern());
				}
				channel.send(exchange.getMessageExchange());
			}
		} else {
			log.log(Level.INFO, "No processing on Unactive MEP");
		}
	}

	/**
	 * Process "in-optional-out" MessageExchange
	 *
	 * @param exchange
	 *            the MessageExchange
	 * @throws MessagingException
	 *             When errors occur during processing
	 */
	private void processInOptionalOut(MessageExchangeWrapper exchange)
			throws MessagingException {
		if (!ackFaultReception(exchange)) {
			Extensions extensions = getExchangeEpJBIExtensions(exchange);
			listener.onJBIMessage(exchange, extensions);

			// Set Status DONE if no Fault or Out messages and ACTIVE Status
			if (exchange.getFault() == null && exchange.getOutMessage() == null
					&& exchange.isActiveStatus()) {
				exchange.setDoneStatus();
			}
		}
	}

	/**
	 * Process "in-only" MessageExchange
	 *
	 * @param exchange
	 *            the MessageExchange
	 * @throws MessagingException
	 *             When errors occur during processing
	 */
	private void processInOnly(final MessageExchangeWrapper exchange)
			throws MessagingException {
		try {
			Extensions extensions = getExchangeEpJBIExtensions(exchange);
			listener.onJBIMessage(exchange, extensions);
		} finally {
			if (exchange.isActiveStatus()) {
				exchange.setDoneStatus();
			}
		}
	}

	/**
	 * Process "in-out" MessageExchange
	 *
	 * @param exchange
	 *            the MessageExchange
	 * @throws MessagingException
	 *             When errors occur during processing
	 *
	 */
	private void processInOut(final MessageExchangeWrapper exchange) {
		Extensions extensions = getExchangeEpJBIExtensions(exchange);
		listener.onJBIMessage(exchange, extensions);
	}

	/**
	 * Process "robust-in-only" MessageExchange
	 *
	 * @param exchange
	 *            the MessageExchange
	 * @throws MessagingException
	 *             When errors occur during processing
	 *
	 */
	private void processRobustInOnly(final MessageExchangeWrapper exchange)
			throws MessagingException {
		Extensions extensions = getExchangeEpJBIExtensions(exchange);
		listener.onJBIMessage(exchange, extensions);

		// Set Status DONE if no Fault or Out messages and ACTIVE Status
		if (exchange.getFault() == null && exchange.isActiveStatus()) {
			exchange.setDoneStatus();
		}
	}

	/**
	 * If the received exchange is a fault response from the consumer,
	 * acknowledge it : status DONE
	 *
	 * @param exchange
	 *            The processed {@link MessageExchange}
	 * @return true if the ack. of the Fault is done; false if no fault is
	 *         present in the exchange
	 * @throws MessagingException
	 */
	private boolean ackFaultReception(final MessageExchangeWrapper exchange)
			throws MessagingException {
		boolean result = false;

		if (exchange.getFault() != null) {
			exchange.setDoneStatus();
			result = true;
		}
		return result;
	}

	private Extensions getExchangeEpJBIExtensions(
			final MessageExchangeWrapper exchange) {
		ServiceEndpoint ep = exchange.getEndpoint();
		Extensions extensions = null;
		if (ep != null) {
			Provides provides = suManager.getProvidesFromEndpoint(ep);
			if (provides != null) {
				extensions = provides.getExtensions();
			}
		}
		return extensions;
	}

	/**
	 * Close the listener. This cause the terminaison of the loop on
	 * DeliveryChannel.accept(), and interrupt this listening thread if the
	 * thread is blocked on DeliveryChannel.accept();
	 */
	public void close() {
		terminate();
		interrupt();
	}

	/**
	 * Start the thread listener ( Thread.start()). The listener waits on
	 * DeliveryChannel.accept(), treats the received MessageExchange, and waits
	 * again on DeliveryChannel.accept().
	 */
	public void listen() {
		start();
	}

	/**
	 * loop on accept Stop on Interrupted Exception
	 */
	public void run() {
		running = true;

		while (running) {

			// receive a message
			MessageExchangeWrapper exchange = null;
			try {
				exchange = new MessageExchangeWrapperImpl(channel.accept());
			} catch (MessagingException e) {
				// search for an interrupted exception
				Throwable cause = e;
				while (cause != null) {
					cause = cause.getCause();
					if (cause instanceof InterruptedException) {
						running = false;
						log.log(Level.SEVERE,
								"Thread interrupted, stop listening.");
					}
				}
				if (running == true) {
					log.log(Level.SEVERE, e.getMessage(), e);
				}
			}

			// forward the message to a worker
			if (exchange != null) {
				if (!isIgnored(exchange)) {
					try {
						process(exchange);
					} catch (Throwable e) {
						log.log(Level.SEVERE, e.getMessage(), e);
						exchange.setError(new Exception(e));
						try {
							channel.send(exchange.getMessageExchange());
						} catch (MessagingException e1) {
							log.log(Level.SEVERE, "Can't send Provider Error "
									+ "MessageExchange to consumer", e);
						}
					}
				}
			}
		}
	}

	/**
	 * Terminate the listening. On the next MessageExchange treated after the
	 * DeliveryChannel.accept(), the listener will stop waiting on
	 * DeliveryChannel.accept().
	 */
	public void terminate() {
		running = false;
	}

	/**
	 * depending on the IgnoredStatus, tells if the exchange has to be ignored.
	 * If the exchange is null, return true.
	 *
	 * @param exchange
	 * @return
	 */
	protected boolean isIgnored(final MessageExchangeWrapper exchange) {
		boolean result = false;

		if (exchange.getMessageExchange() == null) {
			result = true;
		} else {
			if (exchange.isDoneStatus()) {
				result = IgnoredStatusConstants.DONE_IGNORED == configuration
						.getIgnoredStatus()
						|| IgnoredStatusConstants.DONE_AND_ERROR_IGNORED == configuration
								.getIgnoredStatus();
			} else if (exchange.isErrorStatus()) {
				result = IgnoredStatusConstants.ERROR_IGNORED == configuration
						.getIgnoredStatus()
						|| IgnoredStatusConstants.DONE_AND_ERROR_IGNORED == configuration
								.getIgnoredStatus();
			}
		}
		return result;
	}
}
