/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * MessageExchangeSender.java - Initial Developper : ofabre
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.sender;

import java.net.URI;
import java.util.Map;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.common.util.StringHelper;
import org.objectweb.petals.component.common.ExtensionsConstants;
import org.objectweb.petals.component.common.MEPConstants;
import org.objectweb.petals.component.common.su.SimpleServiceUnitManager;
import org.objectweb.petals.component.common.util.CDKExtensions;
import org.objectweb.petals.jbi.descriptor.Consumes;
import org.objectweb.petals.jbi.descriptor.Extensions;

/**
 * A useful class wrapping message exchange sending
 *
 * @author ofabre - EBM Websourcing
 *
 */
public class MessageExchangeSender {

	/**
	 * The component's delivery channel
	 */
	private DeliveryChannel channel;

	/**
	 * The Component's Context
	 */
	private ComponentContext context;

	/**
	 * The component's service unit manager
	 */
	private SimpleServiceUnitManager suManager;

	/**
	 * InOnly Mep shortcut
	 */
	public final static String IN_ONLY = "in-only";

	/**
	 * InOnly Mep shortcut
	 */
	public final static String ROBUST_IN_ONLY = "robust-in-only";

	/**
	 * InOnly Mep shortcut
	 */
	public final static String IN_OUT = "in-out";

	/**
	 * InOnly Mep shortcut
	 */
	public final static String IN_OPTIONAL_OUT = "in-opt-out";

	public MessageExchangeSender(DeliveryChannel channel,
			ComponentContext context, SimpleServiceUnitManager suManager) {
		super();
		this.channel = channel;
		this.context = context;
		this.suManager = suManager;
	}

	/**
	 * Send a message to the endpoint associated to the mapped address. For
	 * exemple, this method can be used by a listener, that listen to messages
	 * from external services (linked to the given address), in order to send
	 * these messages to internal JBI Components.
	 *
	 * @param address
	 *            the distant address of the external consumer service. Must be
	 *            non null and non empty.
	 * @param exchange
	 *            the message exchange to be sent to the internal JBI Component.
	 *            Must be non null.
	 * @return true if the message has been sent (if a timeout occurs, return
	 *         false)
	 * @throws MessagingException
	 *             if the address has no endpoint associated to it, or if an
	 *             error occurs while sending the message through the container
	 */
	public boolean sendSync(final String address, final MessageExchange exchange)
			throws MessagingException {

		if (StringHelper.isNullOrEmpty(address)) {
			throw new IllegalArgumentException(
					"You must provide a non null and non empty address");
		}
		if (exchange == null) {
			throw new IllegalArgumentException(
					"You must provide a non null MessageExchange");
		}

		// Retrieve consumes node for adress
		Consumes consumes = suManager.getConsumesFromAddress(address);
		if (consumes == null) {
			throw new MessagingException("Can't find a consumes node in "
					+ "deployed SUs containing the given address: " + address);
		}

		// retrieve timeout
		CDKExtensions extensions = new CDKExtensions(consumes.getExtensions());

		long timeout = 0;
		if (extensions != null) {
			timeout = Long.parseLong(extensions.getPetalsExtensions()
					.getProperty(ExtensionsConstants.TIMEOUT, "0"));
		}

		// Retrieve possible target endpoint
		ServiceEndpoint ep = null;
		if (consumes.getEndpointName() != null
				&& consumes.getServiceName() != null) {
			ep = context.getEndpoint(consumes.getServiceName(), consumes
					.getEndpointName());
		}

		exchange.setInterfaceName(consumes.getInterfaceName());
		exchange.setService(consumes.getServiceName());
		exchange.setEndpoint(ep);

		return channel.sendSync(exchange, timeout);
	}

	/**
	 * Send a message to the endpoint associated to the mapped address. The
	 * {@link MessageExchange} is created from information given in parameters
	 * (content, attachments, properties, operation and pattern). If no
	 * operation and pattern are specified, it tries to find it in the
	 * "consumes" extensions associated to the given address. If no pattern can
	 * be found, InOnly pattern is used by default.
	 *
	 * @param address
	 *            the distant address of the external consumer service. Must be
	 *            non null and non empty.
	 * @param content
	 *            the {@link Source} content of the "IN" message. Can be null
	 * @param attachments
	 *            a {@link Map} of "IN" message attachments. Can be null or
	 *            empty
	 * @param properties
	 *            a {@link Map} of "IN" message properties. Can be null or empty
	 * @param operation
	 *            the targeted operation. Can be null
	 * @param pattern
	 *            the {@link MessageExchange} pattern. Can be null
	 * @return the {@link MessageExchange} that has been sent
	 * @throws MessagingException
	 *             if there is no consumes node associated to the given address
	 *             or if the message can't be send (error or timeout)
	 */
	public MessageExchange sendSync(final String address, final Source content,
			final Map<String, DataHandler> attachments,
			Map<String, Object> properties, QName operation, URI pattern)
			throws MessagingException {
		boolean sent = false;

		if (StringHelper.isNullOrEmpty(address)) {
			throw new MessagingException(
					"You must provide a non null and non empty address");
		}

		Consumes consumes = suManager.getConsumesFromAddress(address);
		if (consumes == null) {
			throw new MessagingException("Can't find a consumes node in "
					+ "deployed SUs containing the given address: " + address);
		}

		// Get consumes extensions
		CDKExtensions extensions = new CDKExtensions(consumes.getExtensions());

		// Create MessageExchange of the good pattern
		MessageExchange exchange = createMessageExchange(MEPConstants
				.valueOf(pattern), extensions);

		// Set operation to call if exists
		if (operation == null) {
			if (extensions != null) {
				operation = QName.valueOf(extensions.getPetalsExtensions()
						.getProperty(ExtensionsConstants.OPERATION));
			}
		}
		exchange.setOperation(operation);

		// Set "in" MessageExchange
		setInMessage(exchange, content, attachments, properties);

		try {
			sent = sendSync(address, exchange);
		} catch (JBIException e) {
			throw new MessagingException(e);
		}

		if (!sent) {
			throw new MessagingException(
					"Failed to send the mapped message to the delivery channel,"
							+ " sendSync() method returned false");
		}

		return exchange;
	}

	/**
	 * Send a message to the endpoint associated to the mapped address. The
	 * {@link MessageExchange} is created from information given in parameters
	 * (content, attachments, properties).
	 *
	 * @param address
	 *            the distant address of the external consumer service. Must be
	 *            non null and non empty.
	 * @param content
	 *            the {@link Source} content of the "IN" message. Can be null
	 * @param attachments
	 *            a {@link Map} of "IN" message attachments. Can be null or
	 *            empty
	 * @param properties
	 *            a {@link Map} of "IN" message properties. Can be null or empty
	 * @return the {@link MessageExchange} that has been sent
	 * @throws MessagingException
	 *             if there is no consumes node associated to the given address
	 *             or if the message can't be send (error or timeout)
	 */
	public MessageExchange sendSync(final String address, final Source content,
			final Map<String, DataHandler> attachments,
			Map<String, Object> properties) throws MessagingException {
		return sendSync(address, content, attachments, properties, null, null);
	}

	/**
	 * Send a message to the endpoint associated to the mapped address. The
	 * {@link MessageExchange} is created from information given in parameters
	 * (content, attachments).
	 *
	 * @param address
	 *            the distant address of the external consumer service. Must be
	 *            non null and non empty.
	 * @param content
	 *            the {@link Source} content of the "IN" message. Can be null
	 * @param attachments
	 *            a {@link Map} of "IN" message attachments. Can be null or
	 *            empty
	 * @return the {@link MessageExchange} that has been sent
	 * @throws MessagingException
	 *             if there is no consumes node associated to the given address
	 *             or if the message can't be send (error or timeout)
	 */
	public MessageExchange sendSync(final String address, final Source content,
			final Map<String, DataHandler> attachments)
			throws MessagingException {
		return sendSync(address, content, attachments, null, null, null);
	}

	/**
	 * Send a message to the endpoint associated to the mapped address. The
	 * {@link MessageExchange} is created from information given in parameters
	 * (content).
	 *
	 * @param address
	 *            the distant address of the external consumer service. Must be
	 *            non null and non empty.
	 * @param content
	 *            the {@link Source} content of the "IN" message. Can be null
	 * @return the {@link MessageExchange} that has been sent
	 * @throws MessagingException
	 *             if there is no consumes node associated to the given address
	 *             or if the message can't be send (error or timeout)
	 */
	public MessageExchange sendSync(final String address, final Source content)
			throws MessagingException {
		return sendSync(address, content, null, null, null, null);
	}

	/**
	 * Create a message exchange matching the given pattern. If pattern
	 * parameter is null, try to find a specified pattern in "consumes"
	 * extensions. Finaly, if no pattern can be found, an InOnly exchange is
	 * created by default.
	 *
	 * @param mep
	 *            a specific MEP pattern
	 * @param extensions
	 *            "consumes" {@link Extensions}
	 * @return a message exchange matching the given pattern
	 * @throws MessagingException
	 *             if the message exchange can't be created or if the specified
	 *             pattern isn't valid
	 */
	private MessageExchange createMessageExchange(MEPConstants mep,
			CDKExtensions extensions) throws MessagingException {
		if (mep == null) {
			mep = retrievePattern(extensions);
		}
		MessageExchange exchange = null;
		if (mep.equals(MEPConstants.IN_ONLY_PATTERN)) {
			exchange = channel.createExchangeFactory().createInOnlyExchange();
		} else if (mep.equals(MEPConstants.IN_OPTIONAL_OUT_PATTERN)) {
			exchange = channel.createExchangeFactory()
					.createInOptionalOutExchange();
		} else if (mep.equals(MEPConstants.IN_OUT_PATTERN)) {
			exchange = channel.createExchangeFactory().createInOutExchange();
		} else if (mep.equals(MEPConstants.ROBUST_IN_ONLY_PATTERN)) {
			exchange = channel.createExchangeFactory()
					.createRobustInOnlyExchange();
		}
		return exchange;
	}

	private void setInMessage(MessageExchange exchange, Source content,
			Map<String, DataHandler> attachments, Map<String, Object> properties)
			throws MessagingException {
		NormalizedMessage message = exchange.createMessage();
		message.setContent(content);
		if (attachments != null) {
			for (String attachment : attachments.keySet()) {
				message.addAttachment(attachment, attachments.get(attachment));
			}
		}
		if (properties != null) {
			for (String property : properties.keySet()) {
				message.setProperty(property, properties.get(property));
			}
		}
		exchange.setMessage(message, "in");
	}

	/**
	 * Retrieve the MEP from the extensions. If no pattern specified in
	 * extensions, return an InOnly pattern by default
	 *
	 * @param extensions
	 * @return mep specified in extensions
	 * @throws MessagingException
	 */
	private MEPConstants retrievePattern(CDKExtensions extensions)
			throws MessagingException {
		MEPConstants mep = null;
		String patternShortcut = null;
		if (extensions != null) {
			patternShortcut = extensions.getPetalsExtensions().getProperty(
					ExtensionsConstants.MEP);
		}
		if (!StringHelper.isNullOrEmpty(patternShortcut)) {
			if (ROBUST_IN_ONLY.equalsIgnoreCase(patternShortcut)) {
				mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
			} else if (IN_OPTIONAL_OUT.equalsIgnoreCase(patternShortcut)) {
				mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
			} else if (IN_OUT.equalsIgnoreCase(patternShortcut)) {
				mep = MEPConstants.IN_OUT_PATTERN;
			} else if (IN_ONLY.equalsIgnoreCase(patternShortcut)) {
				mep = MEPConstants.IN_ONLY_PATTERN;
			} else {
				throw new MessagingException("MEP pattern specified in "
						+ "consumes extension (" + patternShortcut
						+ ") isn't valid. " + "Valid patterns are : in-only, "
						+ "in-opt-out, robust-in-only, in-out.");
			}
		} else {
			mep = MEPConstants.IN_ONLY_PATTERN;
		}
		return mep;
	}
}
